/* Copyright 2020-2025, Michele Santullo
 * This file is part of memoserv.
 *
 * Memoserv is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Memoserv is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Memoserv.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "block_drag_and_drop.hpp"
#include <QByteArray>
#include <QDataStream>
#include <QMouseEvent>
#include <QMimeData>
#include <QDrag>
#include <QApplication>
#include <QWidget>
#include <cassert>
#include <ciso646>
#include <memory>
#if !defined(NDEBUG)
#	include <iostream>
#endif

namespace duck::widget::detail {
namespace {
constexpr char g_drag_mime[] = "application/x-dndblock";
} //unnamed namespace

BlockDragAndDrop::BlockDragAndDrop (
	QWidget* parent,
	PointToIndexFunc&& func,
	std::function<std::size_t()>&& get_icon_count,
	std::function<const QPixmap&(std::size_t)>&& get_pixmap
) :
	m_point_to_index(std::move(func)),
	m_get_icon_count(std::move(get_icon_count)),
	m_get_pixmap(std::move(get_pixmap)),
	m_parent(parent)
{
	assert(m_parent);
	m_parent->setAcceptDrops(true);
}

void BlockDragAndDrop::mouse_press_event (const QMouseEvent* event) {
	if (event->button() & Qt::LeftButton) {
		assert(not m_dragging);

		const auto [block_index, block_origin] = m_point_to_index(event->position().toPoint());
		if (block_index < 0)
			return;

		const auto icon_index = static_cast<std::size_t>(block_index);
		if (icon_index >= m_get_icon_count())
			return;

		m_start_pos = event->position().toPoint();
		m_dragging = true;
		m_block_index = block_index;
		m_hot_spot = m_start_pos - block_origin;
	}
}

void BlockDragAndDrop::mouse_release_event(const QMouseEvent* event) {
	if (event->button() & Qt::LeftButton) {
		m_dragging = false;
#if !defined(NDEBUG)
		std::cout << "Mouse button released before block's drag&drop\n";
#endif
	}
}

void BlockDragAndDrop::mouse_move_event (const QMouseEvent* event) {
	if (not m_dragging)
		return;

	if (not (event->buttons() & Qt::LeftButton))
		return;

	if ((event->position().toPoint() - m_start_pos).manhattanLength() < QApplication::startDragDistance())
		return;

#if !defined(NDEBUG)
	std::cout << "mouse_move_event() drag&drop started\n";
#endif
	QByteArray item_data;
	QDataStream data_stream(&item_data, QIODevice::WriteOnly);
	data_stream << m_get_pixmap(m_block_index);

	std::unique_ptr<QMimeData> mime_data(new QMimeData);
	mime_data->setData(g_drag_mime, item_data);

	assert(m_block_index >= 0);
	const auto icon_index = static_cast<unsigned int>(m_block_index);
	QDrag* const drag(new QDrag(m_parent));
	drag->setMimeData(mime_data.release());
	drag->setPixmap(m_get_pixmap(icon_index));
	drag->setHotSpot(m_hot_spot);

	switch (drag->exec(Qt::CopyAction | Qt::MoveAction, Qt::CopyAction)) {
	case Qt::MoveAction:
		break;
	case Qt::CopyAction:
		break;
	default:
		break;
	}
	m_dragging = false;
}

void BlockDragAndDrop::drag_enter_event(QDragEnterEvent* event) {
	if (not m_dragging and event->mimeData()->hasFormat(g_drag_mime))
		event->acceptProposedAction();
}

void BlockDragAndDrop::drop_event(QDropEvent* event) {
}
} //namespace duck::widget::detail
